/************************************************************************/
/*									*/
/*			Connect several AVRs over one wire              */
/*		Author: Peter Dannegger					*/
/*									*/
/************************************************************************/
//                      Target: ATmega48
/*
Master:         ATmega644
Address:        none
Data pin:       PB3 = OC0A (transmit) = PCINT11 (receive)
KEY0:           PB4
KEY1:           PB5
KEY2:           PB6
LED0:           PC1
LED1:           PC2
LED2:           PC3

Slave 1:        ATmega48
Address:        1
Data pin:       PB3 = OC2A (transmit) = PCINT3 (receive)
KEY0:           PB0
KEY1:           PB1
LED0:           PC0

Slave 2:        ATtiny261
Address:        2
Data pin:       PB3 = OC1B (transmit) = PCINT11 (receive)
KEY0:           PB2
LED0:           PA6
LED1:           PA7

Data Protocol:
1 * T: o-Bit
2 * T: 1-Bit
4 * T: Sync-Bit
*/

#include <avr/interrupt.h>
#include <avr/power.h>
#include "mydefs.h"

#define F_CPU           8e6                     // 8MHz
#define BIT_CLK         10e3                    // Bit time T = 10kHz
#define PRESCALER       64                      // F_CPU / 64
#define PRESCALER_SET	(1<<CS22)
#define	BIT_TIME	(F_CPU / BIT_CLK / PRESCALER)

#define DATA_SIZE       2                       // 2 Byte

#define	KEY0_PIN        SBIT( PINB,  PB0 )
#define	KEY1_PIN        SBIT( PINB,  PB1 )
#define	LED0            SBIT( PORTC, PC0 )
#define	LED0_DDR        SBIT( DDRC,  PC0 )
#define	LED1            SBIT( PORTC, PC4 )
#define	LED1_DDR        SBIT( DDRC,  PC4 )
#define	LED2            SBIT( PORTC, PC5 )
#define	LED2_DDR        SBIT( DDRC,  PC5 )
#define	DATA            SBIT( PORTB, PB3 )
#define	DATA_DDR        SBIT( DDRB,  PB3 )
#define	DATA_PIN        SBIT( PINB,  PB3 )


/****************************** Transmission ****************************/

uint8_t tx_buff[DATA_SIZE];
uint8_t tx_idx;
#define rx_idx  tx_idx
#define rx_buff tx_buff
#define DATA_INT        SBIT( TIMSK2, OCIE2A )


ISR( TIMER2_COMPA_vect, ISR_NOBLOCK )           // Transmit/Sync interrupt
{
  uint8_t *dp;					// locals for faster access

  if( DATA_DDR ){
    if( tx_idx < DATA_SIZE * 8 ){
      OCR2A += (uint16_t)(BIT_TIME + 0.5);      // 0-bit: 1T
      dp = &tx_buff[tx_idx++ / 8];              // get byte address
      if( *dp & 0x80 )                          // msb first
        OCR2A += (uint16_t)(BIT_TIME + 0.5);    // 1-bit: 2T
      *dp <<= 1;                                // shift bits
      return;
    }
    OCR2A += (uint16_t)(BIT_TIME * 4 + 0.5);    // Sync:  4T
    TCCR2A = 1<<COM2A1 | 1<<COM2A0;             // set on compare
    if( DATA_PIN == 0 )
      return;
  }else{
    if( rx_idx != DATA_SIZE * 8 ){              // receive successful
      rx_idx = 255;                             // ready for start transition
      return;
    }
    PCMSK0 = 0;                                 // disable further receive
  }
  DATA_INT = 0;                                 // disable interrupt
}


void tx_data( uint8_t addr, uint8_t data )
{
  PCMSK0 = 0;    				// disable receive
  DATA_DDR = 0;                                 // input
  DATA = 1;                                     // pullup on
  while( DATA_PIN == 0 );                       // wait until pin high
  OCR2A = TCNT2 + (uint16_t)(BIT_TIME * 4 + 0.5);      // sync: 4T
  TCCR2A = 1<<COM2A0;                           // toggle on compare
  DATA_DDR = 1;                                 // tx output on
  TIFR2 = 1<<OCF2A;                             // clear earlier interrupts
  DATA_INT = 1;                                 // enable tx interrupt
  volatil(tx_idx) = 0;                          // point to first bit
  volatil(tx_buff[0]) = addr;
  volatil(tx_buff[1]) = data;
  while( DATA_INT );                            // until tx done
}


/****************************** Receiving *******************************/

ISR( PCINT0_vect, ISR_NOBLOCK )                 // Receive interrupt
{
  uint8_t *dp;					// locals for faster access
  uint8_t i;                                	// locals for faster access
  uint8_t t;

  t = TCNT2;
  if( rx_idx >= DATA_SIZE * 8 ){                // on start or stop bit
    rx_idx = 0;                                 // point to first bit
  }else{
    dp = &rx_buff[rx_idx++ / 8];                // get byte address
    i = *dp << 1;                               // shift bits
    if((uint8_t)(OCR2A - t) < (uint8_t)(BIT_TIME * 1.4 + 0.5))  // if pulse > 1.4T
      i++;                                      // msb first
    *dp = i;
  }
  OCR2A = t + (uint8_t)(BIT_TIME * 2.8 + 0.5);  // detect 4T
}


#define	rx_busy() PCMSK0                        // 0 = data received


void rx_init( void )
{
  DATA_DDR = 0;                                 // input
  DATA = 1;                                     // pullup on
  OCR2A = TCNT2 + (uint16_t)(BIT_TIME * 2.8 - 0.5);    // detect 4T
  volatil(rx_idx) = 255;                            // detect sync transition
  TIFR2 = 1<<OCF2A;                             // clear earlier interrupts
  DATA_INT = 1;                                 // enable sync interrupt
  PCMSK0 = 1<<PCINT3;
  PCIFR = 1<<PCIF0;                             // clear earlier interrupts
  PCICR = 1<<PCIE0;                             // pin change interrupt
}


/****************************** PWM *************************************/

ISR( TIMER2_COMPB_vect )
{
  LED1 = 1;                                     // LED off
}


ISR( TIMER2_OVF_vect )
{
  if( OCR2B != 0xFF )
    LED1 = 0;                                   // LED on
}


void pwm_init( void )
{
  OCR2B = 0x1F;                                 // 12.5%
  TIMSK2 = 1<<OCIE2B | 1<<TOIE2;
  TCCR2A = 1<<COM2A1 | 1<<COM2A0;               // set data pin on compare
  TCCR2B = PRESCALER_SET;                       // start T2
}


/****************************** Debounce ********************************/

uint8_t key_state;			// debounced and inverted key state:
					// bit = 1: key pressed
uint8_t key_press;			// key press detect


void debounce( void )
{
  static uint8_t ct0 = 0xFF, ct1 = 0xFF;	// 8 * 2bit counters
  uint8_t i;

  i = ~PINB;				// read keys (low active)
  i ^= key_state;			// key changed ?
  ct0 = ~( ct0 & i );			// reset or count ct0
  ct1 = ct0 ^ (ct1 & i);		// reset or count ct1
  i &= ct0 & ct1;			// count until roll over ?
  key_state ^= i;			// then toggle debounced state
  key_press |= key_state & i;		// 0->1: key press detect
}


uint8_t get_key_press( uint8_t key_mask )
{
  key_mask &= key_press;		// read key(s)
  key_press ^= key_mask;		// clear key(s)
  return key_mask;
}


/****************************** Main ************************************/

int main( void )
{
  clock_prescale_set( clock_div_1 );
  PORTB = 0xFF;					// input with pullup
  PORTC = 0xFF;					// input with pullup
  PORTD = 0xFF;					// input with pullup
  LED0 = 0;
  LED0_DDR = 1;                                 // Output
  LED1 = 0;
  LED1_DDR = 1;                                 // Output
  pwm_init();

  sei();
  for(;;){
    rx_init();                                  // start receive
    while( rx_busy() );
    if( volatil(rx_buff[0]) == 1 ){             // if slave 1 addressed
      LED0 = volatil(rx_buff[1]) & 1;
      tx_data( 1, KEY0_PIN | KEY1_PIN<<1 );     // answer from slave 1
      debounce();
      if( get_key_press( 1<<PB0 ))
        OCR2B++;
      if( get_key_press( 1<<PB1 ))
        OCR2B--;
    }
  }
}
